/*
  File: ConditionVariable.java

  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.

  History:
  Date       Who                What
  11Jun1998  dl               Create public version


  This is a slightly adapted and simplified version for the use in
  the operating systems lecture at University of Siegen, Germany.
  Source of the original code: http://gee.cs.oswego.edu/dl/classes/
  EDU/oswego/cs/dl/util/concurrent/intro.html
*/

package BSsync;

/**
 * This class implements condition variables similar to POSIX pthreads.
 * If you restrict yourself to Locks and Conditions, you can
 * use most of your favorite constructions. Don't randomly mix them
 * with synchronized methods or blocks though.
 * <p>
 * Method names and behavior are as close as is reasonable to
 * those in POSIX.
 * <p>
 * <b>Sample Usage.</b> Here is a full version of a bounded buffer
 * that implements the BoundedChannel interface, written in
 * a style reminscent of that in POSIX programming books.
 * <pre>
 * class CVBuffer implements BoundedChannel {
 *   private final Lock mutex;
 *   private final Condition notFull;
 *   private final Condition notEmpty;
 *   private int count = 0;
 *   private int takePtr = 0;     
 *   private int putPtr = 0;
 *   private final Object[] array;
 * 
 *   public CVBuffer(int capacity) { 
 *     array = new Object[capacity];
 *     mutex = new Lock();
 *     notFull = mutex.newCondition();
 *     notEmpty = mutex.newCondition();
 *   }
 * 
 *   public int capacity() { return array.length; }
 * 
 *   public void put(Object x) {
 *       mutex.lock();
 *       while (count == array.length) {
 *         notFull.await();
 *       }
 *       array[putPtr] = x;
 *       putPtr = (putPtr + 1) % array.length;
 *       ++count;
 *       notEmpty.signal();
 *       mutex.unlock();
 *   }
 * 
 *   public Object take() {
 *       Object x = null;
 *       mutex.lock();
 *       while (count == 0) {
 *         notEmpty.await();
 *       }
 *       x = array[takePtr];
 *       array[takePtr] = null;
 *       takePtr = (takePtr + 1) % array.length;
 *       --count;
 *       notFull.signal();
 *       mutex.unlock();
 *       return x;
 *   }
 * }
 *
 * </pre>
 * @see Lock
 **/

public interface Condition
{
	/** 
	 * Wait for notification. This operation at least momentarily
	 * releases the mutex. The mutex is always acquired back before
	 * the method returns.
	 **/
	public void await();
	
	/** 
	 * Notify a waiting thread.
	 * If one exists, a thread will return from await.
	 **/
	public void signal();

	/** Notify all waiting threads **/
	public void signalAll();
}
