/*
  File: Mutex.java

  Originally written by Doug Lea and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.
  Thanks for the assistance and support of Sun Microsystems Labs,
  and everyone contributing, testing, and using this code.

  History:
  Date       Who                What
  11Jun1998  dl               Create public version


  This is a slightly adapted and simplified version for the use in
  the operating systems lecture at University of Siegen, Germany.
  Source of the original code: http://gee.cs.oswego.edu/dl/classes/
  EDU/oswego/cs/dl/util/concurrent/intro.html
*/

package BSsync;

/**
 * A simple non-reentrant mutual exclusion lock.
 * The lock is free upon construction. Each acquire gets the
 * lock, and each release frees it. Releasing a lock that
 * is already free has no effect. 
 * <p>
 * This implementation makes no attempt to provide any fairness
 * or ordering guarantees.
 * @see Semaphore
**/

public class Lock
{
	/** Set this variable to true for debugging output */
	public static boolean verbose = false;

	/** The lock status **/
	protected boolean inuse;

	/** The mutex name (for debugging) **/
	String name;
	static int seqno = 1;

	/** Constructor for a mutex lock. */
	public Lock()
	{
		inuse = false;
		name = "Lock-" + seqno;
		seqno++;
	}

	/**
	 * Constructor for a mutex lock with a given name. 
	 * The name is only used for debugging output.
	 * @param name name of the mutex
	*/
	public Lock(String name)
	{
		inuse = false;
		this.name = name;
	}

	/** Acquire the mutex lock. 
	 ** If the mutex is already locked, the calling thread is blocked
	 ** until the mutex is available.
	 */
	public synchronized void lock() {
		try {
			boolean wasInUse = inuse;
			if (verbose && inuse) {
				System.out.println(Thread.currentThread().getName()
								   + ": " + name 
								   + ".lock() : warte ...");
		  		System.out.flush();
			}
			while (inuse)
				wait();
			inuse = true;
			if (verbose) {
				if (wasInUse) {
					System.out.println(Thread.currentThread().getName()
									   + ": " + name + ".lock() : fertig");
				}
				else {
					System.out.println(Thread.currentThread().getName()
									   + ": " + name + ".lock()");
				}
		  		System.out.flush();
			}
		}
		catch (InterruptedException ex) {
			System.err.println("Interrupt, exiting ...");
			System.exit(1);
		}
	}
	
	/** Release the mutex lock. */
	public synchronized void unlock()  {
		if (verbose) {
			System.out.println(Thread.currentThread().getName()
							   + ": " + name + ".unlock()");
		  	System.out.flush();
		}
		inuse = false;
		notify();
	}

	/** Create and return a new condition variable associated with this lock.
	 * In normal use, the lock should be one that is used for <em>all</em>
	 * synchronization of the object using the Condition. Generally, 
	 * to prevent nested monitor lockouts, this
	 * object should not use any native Java synchronized blocks.
	 **/
	public Condition newCondition() {
		return new CondVar(this);
	}

	/** Create and return a new condition variable associated with this lock.
	 * The name is just used for debugging output.
	 * In normal use, the lock should be one that is used for <em>all</em>
	 * synchronization of the object using the Condition. Generally, 
	 * to prevent nested monitor lockouts, this
	 * object should not use any native Java synchronized blocks.
	 * @param name name of the Condition object.
	 **/
	public Condition newCondition(String name) {
		return new CondVar(this, name);
	}
}


class CondVar implements Condition
{
	/** The mutex **/
	protected final Lock mutex_;

	/** The name of the condition variable (for debugging) **/
	String name;
	static int seqno = 1;

	/** 
	 * Create a new CondVar that relies on the given mutual
	 * exclusion lock. 
	 * @param mutex A non-reentrant mutual exclusion lock.
	 * In normal use, the mutex should be one that is used for <em>all</em>
	 * synchronization of the object using the CondVar. Generally, 
	 * to prevent nested monitor lockouts, this
	 * object should not use any native Java synchronized blocks.
	 **/
	CondVar(Lock mutex) {
		mutex_ = mutex;
		name = "Condition-" + seqno;
		seqno++;
	}

	/** 
	 * Create a new named CondVar that relies on the given mutual
	 * exclusion lock. 
	 * The name is just used for debugging output.
	 * @param mutex A non-reentrant mutual exclusion lock.
	 * In normal use, the mutex should be one that is used for <em>all</em>
	 * synchronization of the object using the CondVar. Generally, 
	 * to prevent nested monitor lockouts, this
	 * object should not use any native Java synchronized blocks.
	 * @param name name of the CondVar.
	 **/
	CondVar(Lock mutex, String name) {
		mutex_ = mutex;
		this.name = name;
	}
	
	
  /** 
   * Wait for notification. This operation at least momentarily
   * releases the mutex. The mutex is always acquired back before
   * the method returns.
   **/
  public void await() {
	  synchronized (this) {
		  if (Lock.verbose) {
			  System.out.println(Thread.currentThread().getName()
								 + ": " + name + ".await() : gebe Lock auf und warte ...");
			  System.out.flush();
		  }
		  mutex_.unlock();
		  try {
			  wait(); 
		  }
		  catch (InterruptedException ex) {
			  System.err.println("Interrupt, exiting ...");
			  System.exit(1);
		  }
      }
	  /* It is important to give up the monitor here, otherwise, we
		 run into deadlocks! */

	  if (Lock.verbose) {
		  System.out.println(Thread.currentThread().getName()
							 + ": " + name + ".await() : signalisiert, hole Lock wieder ...");
		  System.out.flush();
	  }
	  mutex_.lock();
	  if (Lock.verbose) {
		  System.out.println(Thread.currentThread().getName()
							 + ": " + name + ".await() : fertig");
		  System.out.flush();
	  }
  }

  /** 
   * Notify a waiting thread.
   * If one exists, a thread will return from await.
   **/
  public synchronized void signal() {
	  if (Lock.verbose) {
		  System.out.println(Thread.currentThread().getName()
							 + ": " + name + ".signal()");
		  System.out.flush();
	  }
	  notify();
  }

  /** Notify all waiting threads **/
  public synchronized void signalAll() {
	  if (Lock.verbose) {
		  System.out.println(Thread.currentThread().getName()
							 + ": " + name + ".signalAll()");
		  System.out.flush();
	  }
	  notifyAll();
  }
}
